# Default tzk config
# <<GENERATION_DETAILS>>

# This config file is live, unrestricted Python.
# You can add your own builders or logic here if necessary.
# tzk's package directory is on the PYTHONPATH when this module is imported,
# so if you want to import a function from, e.g., tzk.util,
# just say 'from tzk.util import whatever'.

#------------------------------------------------------------------------------#

# Imports. Don't remove this or you won't be able to configure builders.
from tzk import builders

# Name of the subfolder containing your wiki data and its tiddlywiki.info file.
wiki_folder = "wiki"


### COMMITTING ####
# Default commit message to use with 'tzk commit'.
# You can always use 'tzk commit -m' to use a different message on the fly.
commit_message = "checkpoint"

# Git remote to push changes to when you run 'tzk commit'.
# If you never want to push changes, set this to the empty string ("").
commit_remote = ""
#commit_remote = "origin"

# Uncomment if you want to abort 'tzk commit' if you're not on a specific branch.
#commit_require_branch = "master"


### LISTENING ###
# Port to listen on. If you specify 8080 for example, you'll edit your wiki by going to
# http://localhost:8080 in your browser.
listen_port = 8080

# Uncomment if you want to require HTTP basic authentication when serving your wiki.
# **WARNING**: this is NOT secure for use over the open Internet or all but the
# simplest local networks, as the password is sent in the clear. For good
# security, you need to add a reverse proxy on top that secures the connection via SSL.
#listen_username = "my_user"
#listen_password = "my_password"


### BUILD ###
# Filter for tiddlers that should be included in a public export of the wiki.
_public_export_filt = r"""
    [is[system]]
    [tag[Public]]
    -[[$:/plugins/tiddlywiki/tiddlyweb]]
    -[[$:/plugins/tiddlywiki/filesystem]]
    -[prefix[$:/temp]]
    -[prefix[$:/state]]
    -[prefix[$:/sib/StorySaver/saved]]
    -[prefix[$:/checkify/]]
    -[[$:/config/zettelkasten/Build/KillPhrases]]
"""

# Each "product" is a different deliverable or target. A product can be built
# by running 'tzk build PRODUCT'. You can include any number of products,
# each of which can use any number of "builders", micro-build steps that
# perform a small part of the build. The configured builders are run in sequence.
#
# See the "Builders" section of the tzk documentation
# for details on the available builders and their parameters.
#
# If you want to do something that's not covered by the builders shipped with tzk,
# you can write your own, or you can run arbitrary shell commands
# using a 'builders.shell("my shell command here"),' builder.
products = {
    # The default configuration contains a single product for building a public wiki;
    # use 'tzk build public' to build it. You can add as many products as you want.
    'public': [
        # Create a temporary folder to hold the public wiki as we work on it.
        builders.new_output_folder(),
        # Pull the tiddlers matching the filter we defined above from our private wiki
        # and put them in a new public wiki in the temporary folder.
        builders.export_public_tiddlers(export_filter=_public_export_filt),
        # Replace the names of people who don't have public tiddlers with their initials.
        builders.replace_private_people(),
        # Change the 'text' field of these tiddlers to new values.
        builders.set_tiddler_values(text={
            '$__config_sib_CurrentEditionPublicity.tid': 'public',
            '$__config_sib_IsPublicEdition.tid': 'false',
            '$__config_DefaultSidebarTab.tid': '$:/sib/SideBar/Explore',
            '$__DefaultTiddlers.tid': 'PublicHomepage',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_close-all.tid': 'show',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_control-panel.tid': 'hide',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_home.tid': 'show',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_more-page-actions.tid': 'show',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_new-tiddler.tid': 'hide',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_permaview.tid': 'show',
            '$__config_PageControlButtons_Visibility_$__core_ui_Buttons_tag-manager.tid': 'hide',
            '$__config_PageControlButtons_Visibility_$__sib_Buttons_NewSource.tid': 'hide',
            '$__config_PageControlButtons_Visibility_$__sib_Buttons_ReadingInbox.tid': 'hide',
            '$__config_ViewToolbarButtons_Visibility_$__core_ui_Buttons_close-others.tid': 'show',
            '$__config_ViewToolbarButtons_Visibility_$__core_ui_Buttons_edit.tid': 'hide',
            '$__config_ViewToolbarButtons_Visibility_$__core_ui_Buttons_info.tid': 'hide',
            '$__config_ViewToolbarButtons_Visibility_$__core_ui_Buttons_permalink.tid': 'show',
            '$__config_ViewToolbarButtons_Visibility_$__sib_Buttons_CopyPublicZettelkastenLink.tid': 'hide',
            '$__config_ViewToolbarButtons_Visibility_DoCopyTitleReference.tid': 'hide',
        }),
        # Fail the build if kill phrases (things we definitely don't want to make
        # public, configured within the wiki) are found in the build.
        builders.check_for_kill_phrases(),
        # Save images to separate files.
        builders.save_attachments_externally(),
        # Create a single HTML file from the public wiki, externalizing the images
        # as we do, and copy it and the extimages folder to the output/public_wiki
        # folder inside our private wiki.
        builders.compile_html_file(externalize_attachments=True),
    ],
    # If you want a second product, add it like this:
    #'secondproduct': [
    #    builders.new_output_folder(),
    #    ... and so on ...
    #],
}



"""
util.py - miscellaneous utility functions
"""
from contextlib import contextmanager
import json
import os
from pathlib import Path
import shutil
import sys
from typing import Any, Callable, Dict, NoReturn


TZK_VERSION = "0.1.2"


class BuildError(Exception):
    pass


def alter_tiddlywiki_info(
        info_path: Path,
        edit_func: Callable[[Dict[str, Any]], Dict[str, Any]]) -> None:
    "Change a tiddlywiki.info (or other json file) in an arbitrary manner."
    with info_path.open("r") as f:
        tinfo = json.load(f)
    tinfo = edit_func(tinfo)
    with info_path.open("w") as f:
        json.dump(tinfo, f, indent=2)


def fail(msg: str, exit_code: int = 1) -> NoReturn:
    "Print message to stderr and quit with exit code 1."
    print(msg, file=sys.stderr)
    sys.exit(exit_code)


def numerize(number: int, singular: str, plural: str = None) -> str:
    "Render a string in the singular or plural as appropriate."
    if plural is None:
        plural = singular + 's'

    if number == 1:
        return singular
    else:
        return plural


@contextmanager
def pushd(directory: str):
    """
    Change directory into the directory /directory/ until the end of the with-block,
    then return to previous directory.
    """
    old_directory = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(old_directory)

def require_dependencies() -> None:
    """
    Raise an exception if dependencies of tzk aren't available.
    """
    if shutil.which("npm") is None:
        fail("npm is not available. "
                "Please install NPM and make it available on your PATH.\n"
                "https://docs.npmjs.com/downloading-and-installing-node-js-and-npm")

    if shutil.which("git") is None:
        fail("Git is not available. "
                "Please install Git and make it available on your PATH.\n"
                "https://git-scm.com/book/en/v2/Getting-Started-Installing-Git")



import subprocess
from typing import Sequence


def exec(*args: str) -> None:
    "Execute a Git command, raising CalledProcessError if the exit code is nonzero."
    subprocess.check_call(["git", *args])


def rc(*args: str) -> int:
    "Execute a Git command, returning the exit code."
    return subprocess.call(["git", *args])


def read(*args: str) -> str:
    "Execute a Git command, returning the output as a string."
    return subprocess.check_output(["git", *args], text=True).strip()



"""
*Builders* are small executable chunks that together can be linked into a useful build
process.

Builders are decorated with :func:`tzk_builder`, which causes them to be
lazy-evaluated: that is, when they're initially called in the configuration,
instead of running the function and returning its result, a zero-argument
function with all of the arguments wrapped up is returned, to be run at a later
time. This allows the configuration file to be read at any time to retrieve
information about the defined products without actually running any build steps.
"""

from collections.abc import Mapping
from contextlib import contextmanager
import functools
import os
from pathlib import Path
import re
import shutil
import subprocess
import tempfile
from typing import Callable, Dict, List, Optional, Set, Sequence, Tuple

from tzk import git
from tzk import tw
from tzk.util import alter_tiddlywiki_info, BuildError, pushd


def tzk_builder(func):
    """
    Decorator which makes a function lazy-evaluable: that is, when it's
    initially called, it returns a zero-argument lambda with the arguments
    initially passed wrapped up in it. Calling that lambda has the effect
    of executing the builder.

    We use this in tzk to allow the user to use function calls in her config
    to define the build steps, while not requiring her to write a bunch of
    ugly and confusing lambda:'s in the config. The functions that will be called
    are prepared during the config and executed later.
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        my_args = args
        my_kwargs = kwargs
        @functools.wraps(new_func)
        def inner():
            func(*my_args, **my_kwargs)
        return inner
    return new_func


def stop(message: str) -> None:
    "Stop the build due to an error condition."
    raise BuildError(message)


def info(message: str) -> None:
    "Print information about this build step to the console."
    print(message)


# Global state available to all builders.
build_state = {}


@tzk_builder
def say_hi(username: str) -> None:
    """
    Say hi to the specified user.

    This function is intended for testing. It will normally succeed and print
    the provided *username*, but if you give "Jeff" as the username, the step
    will fail, and if you give "General Failure" as the username, it will
    cause an unhandled exception.

    :param username: The name of the person to say hi to.
    """
    if username.lower() == 'general failure':
        raise Exception("")
    elif username.lower() == 'jeff':
        # https://twitter.com/yephph/status/1249246702126546944
        stop("Sorry, the name Jeff does not work with our database schema.")
    else:
        print(f"Hello {username}!")


@tzk_builder
def require_branch(branchname: str) -> None:
    """
    Require a specific Git branch to be checked out.

    If the branch isn't checked out, the build will fail immediately. This may
    be helpful if you want to be sure you aren't accidentally building a wiki
    from tentative changes or an old version.

    :param branchname: The name of the branch that must be checked out.
    """
    if git.read("branch", "--show-current") != branchname:
        stop(f"You may only run this build from the {branchname} branch.")


@tzk_builder
def require_clean_working_tree() -> None:
    """
    Require the working tree of the Git repository to be clean.

    If there are any unstaged changes to existing files or staged changes,
    the build will fail immediately.

    For the standard build process, it is not necessary for the working tree
    to be clean. However, if you use any custom build steps that compare history,
    or you simply want to ensure that you always have a recent checkpoint in your
    local version whenever you publish another version, this may be a useful
    requirement.
    """
    pleasecommit = "Please commit or stash them before publishing (try 'tzk commit')."
    if git.rc("diff-index", "--quiet", "--cached", "HEAD", "--") != 0:
        stop(f"There are staged changes. {pleasecommit}")
    if git.rc("diff-files", "--quiet") != 0:
        stop(f"There are uncommitted changes. {pleasecommit}")


@tzk_builder
def new_output_folder():
    """
    Create a new temporary folder to hold intermediate steps of the product being built.
    
    The path to this temporary folder will be stored in the ``public_wiki_folder``
    key of the ``builders.build_state`` dictionary. Future build steps can access
    the work in progress here. A cleaner is registered to delete this folder
    when all steps complete, so any finished product should be copied out by a
    later build step once it is complete.
    """
    assert 'public_wiki_folder' not in build_state
    build_state['public_wiki_folder'] = tempfile.mkdtemp()

def new_output_folder_cleaner():
    if 'public_wiki_folder' in build_state:
        shutil.rmtree(build_state['public_wiki_folder'])
new_output_folder.cleaner = new_output_folder_cleaner


@tzk_builder
def export_public_tiddlers(export_filter: str) -> None:
    """
    Export specified tiddlers to a new wiki in the temporary build folder.

    :func:`new_output_folder()` must be run prior to this builder.

    :param export_filter: A TiddlyWiki filter describing the tiddlers to be selected
                          for inclusion in the new wiki.
    """
    assert 'public_wiki_folder' in build_state, "new_output_folder builder must run first"
    tw.exec((
        ("savewikifolder", build_state['public_wiki_folder'], export_filter),
    ))


def _find_kill_phrases(phrases: Set[str]):
    """
    Search all tiddlers in the public_wiki_folder for the specified kill phrases.
    """
    regexes = [re.compile(phrase) for phrase in phrases]
    failures = []
    tid_files = (Path(build_state['public_wiki_folder']) / "tiddlers").glob("**/*.tid")

    for tid_file in tid_files:
        with tid_file.open() as f:
            for line in f:
                for regex in regexes:
                    if re.search(regex, line):
                        failures.append((regex, str(tid_file), line))
    
    return failures


@tzk_builder
def check_for_kill_phrases(kill_phrase_file: str = None) -> None:
    """
    Fail the build if any of a series of regexes matches a tiddler's source in the temp wiki.

    The temp wiki should be created first
    using the :func:`export_public_tiddlers()` builder.

    The kill phrases are Python-format regular expressions and may be configured
    within the wiki, currently in $:/sib/gui/KillPhrases.

    :param kill_phrase_file: The path from the source wiki's root directory to the
                             config tiddler containing kill phrases. In the default
                             Zettelkasten edition, this is
                             "tiddlers/$__config_zettelkasten_Build_KillPhrases.tid";
                             if you change the way paths are determined, you can give
                             a different path here.
    """
    assert 'public_wiki_folder' in build_state, "new_output_folder builder must run first"
    if kill_phrase_file is None:
        kill_phrase_file = "tiddlers/$__config_zettelkasten_Build_KillPhrases.tid"

    kill_phrases = set()
    with open(kill_phrase_file) as f:
        reading = False
        for line in f:
            if not line.strip():
                reading = True
            if line.strip() and reading:
                kill_phrases.add(line)

    failures = _find_kill_phrases(kill_phrases)
    if failures:
        result = ["Kill phrases were found in your public wiki:"]
        for failed_regex, failed_file, failed_line in failures:
            trimmed_file = failed_file.replace(build_state['public_wiki_folder'], '')
            result.append(f"'{failed_regex.pattern}' matched file {trimmed_file}:\n"
                          f"    {failed_line.strip()}")
        stop('\n'.join(result))


@tzk_builder
def save_attachments_externally(attachment_filter: str = "[is[image]]",
                                extimage_folder: str = "extimage") -> None:
    """
    Save embedded files in the temp wiki into an external folder.

    The temp wiki should be created first
    using the :func:`export_public_tiddlers()` builder.

    Note that this builder **does not finish externalizing images**.
    It saves the images outside the wiki,
    but it does not change the ``_canonical_uri`` and ``text`` fields
    on each image tiddler to point to this new location.
    For the latter step, use the
    ``externalize_attachments``, ``attachment_filter``, and ``canonical_uri_template``
    parameters to the :func:`compile_html_file` step.

    :param attachment_filter: The tiddlers to be saved to the external folder;
                              by default, ``[is[image]]``.
    :param extimage_folder:   The name of the external folder to save to. This must
                              be the default of ``extimage`` to work with the default
                              ``canonical_uri_template``
                              in the :func:`compile_html_file()` builder,
                              but you can use a different name and a different
                              canonical URI template if you prefer.
    """
    assert 'public_wiki_folder' in build_state

    tw.exec(
        (
            ("savetiddlers", attachment_filter, extimage_folder),
        ),
        base_wiki_folder=build_state['public_wiki_folder']
    )


@tzk_builder
def compile_html_file(
        wiki_name: str = "index.html",
        output_folder: str = "output/public_site/",
        overwrite: bool = True,
        externalize_attachments: bool = False,
        attachment_filter: str = "[is[image]]",
        canonical_uri_template: str = "$:/core/templates/canonical-uri-external-image",
    ) -> None:
    """
    Compile a single HTML file from the temp wiki.
    
    Before compiling an HTML file,
    you should create a temp wiki using the :func:`export_public_tiddlers()` builder,
    then run any other build steps you want to use
    to make changes to the wiki being built.
    Once you compile the HTML file,
    it will be copied out to an output location outside the temp folder,
    and the content can no longer be changed through tzk.

    :param wiki_name:     The filename of the single-file wiki to create.
                          Default ``index.html``.
    :param output_folder: The path to the folder
                          where the single-file wiki and any externalized attachments
                          will be placed, relative to the private wiki's root directory.
                          Default ``output/public_site``.
    :param overwrite:     If the ``output_folder`` already exists,
                          should we overwrite any files in it with the same name?
                          Default True.
    :param externalize_attachments: If True, update tiddlers that match
                                    the ``attachment_filter`` parameter to point to
                                    external versions of their content.
                                    Only useful if you have previously run
                                    the ``save_attachments_externally`` builder.
                                    Default False.
    :param attachment_filter:       If externalizing attachments,
                                    which tiddlers should be externalized?
                                    Default ``[is[image]]``.
    :param canonical_uri_template:  What TiddlyWiki template should be used
                                    to determine the new content
                                    of the ``_canonical_uri`` field?
                                    Default ``$:/core/templates/canonical-uri-external-image``.
                                    If you're not sure what this is, don't change it.
    """
    assert 'public_wiki_folder' in build_state

    commands: List[Tuple[str, ...]] = []
    if externalize_attachments:
        commands.extend([
            ("setfield", attachment_filter, "_canonical_uri",
             canonical_uri_template, "text/plain"),
            ("setfield", attachment_filter, "text", "", "text/plain"),
        ])
    commands.append(("render", "$:/core/save/all", wiki_name, "text/plain"))

    tw.exec(commands, base_wiki_folder=build_state['public_wiki_folder'])
    if os.path.exists(output_folder) and not overwrite:
        stop(f"The output folder '{os.path.abspath(output_folder)}' already exists. "
             f"(To overwrite any files existing in the output folder, "
             f"set overwrite = True for this builder.)")

    shutil.copytree(
        Path(build_state['public_wiki_folder']) / "output",
        Path(output_folder),
        dirs_exist_ok=True
    )
    info(f"Successfully copied built output to {os.path.abspath(output_folder)}.")


def _private_people_replacement_table(
        initialer: Callable[[str], str] = None) -> Dict[str, str]:
    "Build table of private people and their transformed initials."

    def _initials_from_tiddler_name(name: str) -> str:
        m = re.match(r"^(?:Mr|Ms|Mx|The)(?P<camel_case_name>.*?)\.tid", name)
        assert m
        return '.'.join(i for i in m.group('camel_case_name') if i.isupper()) + '.'
    
    if initialer is None:
        initialer = _initials_from_tiddler_name

    tiddlers = (Path.cwd() / "tiddlers").glob("**/*.tid")
    person_tiddlers = (i for i in tiddlers if re.match("^(Mr|Ms|Mx|The)", i.name))
    private_person_tiddlers = []
    for pt in person_tiddlers:
        with pt.open() as f:
            for line in f:
                if line.startswith("tags:"):
                    if re.search(r'\bPublic\b', line):
                        # If there's a tags line in the file and it contains the
                        # Public tag, we skip it.
                        break
                    else:
                        # Otherwise, if there's a tags line in the file and it
                        # doesn't contain the Public tag, it's private.
                        private_person_tiddlers.append(pt)
                        break
                if not line.strip():
                    # And if there's no tags line in the file at all,
                    # it's private by default.
                    private_person_tiddlers.append(pt)
                    break
    return {
        i.name.replace('.tid', ''): initialer(i.name)
        for i in private_person_tiddlers
    }


@tzk_builder
def replace_private_people(initialer: Callable[[str], str] = None) -> None:
    """
    Replace the names of people who are not marked Public with their initials.

    If you have lots of PAO (People, Animals, and Organizations) in your Zettelkasten
    and many of them are personal friends,
    you might prefer not to make everything you said about them
    easily searchable on the internet.
    This is more challenging than simply not marking their tiddlers public,
    since there will also probably be links to their tiddlers in other tiddlers --
    and those links contain their full names, if you put them in the title.

    This builder replaces all links, bracketed or WikiCamelCase,
    to the names of all people *not* tagged Public
    with the initials suggested by their CamelCase titles
    (e.g., MsJaneDoe becomes J.D.). The links point to the tiddler ``PrivatePerson``,
    which explains this process.

    :param initialer: If you don't like the way that initials
                      are generated from tiddler filenames by default,
                      you can customize it by passing a callable
                      that takes one string argument
                      (a tiddler filename without the full path, e.g., ``MsJaneDoe.tid``)
                      and returns a string to be considered the "initials" of that person.
    """
    assert 'public_wiki_folder' in build_state

    replacement_table = _private_people_replacement_table(initialer)
    tid_files = (Path(build_state['public_wiki_folder']) / "tiddlers").glob("**/*.tid")

    for tiddler in tid_files:
        dirty = False
        with tiddler.open() as f:
            lines = f.readlines()
        for idx, line in enumerate(lines):
            for replace_person, replace_initials in replacement_table.items():
                if replace_person in line:
                    if '|' + replace_person + ']]' in line:
                        # link with the person as the target only;
                        # beware that you might have put something private in the text
                        lines[idx] = line.replace(replace_person,
                                                  'PrivatePerson')
                    elif '[[' + replace_person + ']]' in line:
                        # link with the person as the target and text
                        lines[idx] = line.replace(replace_person,
                                                  replace_initials + '|PrivatePerson')
                    else:
                        # camel-case link or unlinked reference in text;
                        # or spurious substring, so rule that out with the '\b' search
                        lines[idx] = re.sub(
                            r"\b" + re.escape(replace_person) + r"\b",
                            f'<<privateperson "{replace_initials}">>',
                            line
                        )
                    dirty = True
        if dirty:
            with tiddler.open("w") as f:
                f.writelines(lines)


def _set_fields(mappings: Dict[str, str],
                editing_func: Callable[[Path, List[str], str], None]) -> None:
    """
    Read the text of a .tid file into memory and call an editing_func on it
    in order to modify some of its fields.

    :param mappings: Mapping of tiddler filenames to new values of some field
                     (this function is field-agnostic; the editing_func should be
                     aware of what field is to be edited).
    :param editing_func: A function which will make the changes.

    The editing_func is called once for each item in the mapping
    and receives three arguments:

    :param tiddler_path: The full path to the tiddler on the filesystem.
                         The editing_func should write changes back to this file,
                         if any are appropriate.
    :param tiddler_lines: A sequence of strings, each one line of the original file.
    :param new_text: The new field value specified for this tiddler in the provided
                     mapping dict of tiddler name/new value.
    """

    for tiddler, new_text in mappings.items():
        tiddler_path = (Path(build_state['public_wiki_folder']) / "tiddlers" / tiddler)
        try:
            with tiddler_path.open("r") as f:
                tiddler_lines = f.readlines()
        except FileNotFoundError:
            stop(f"File {tiddler_path} not found. "
                 f"(Did you forget to end the name with '.tid'?)")

        editing_func(tiddler_path, tiddler_lines, new_text)


def _set_text_field(mappings: Dict[str, str]) -> None:
    "Set the 'text' field to a new value in each pair of tiddler name/value."
    def editor(tiddler_path: Path, tiddler_lines: Sequence[str], new_text: str) -> None:
        if not str(tiddler_path).endswith('.tid'):
            # will be a separate meta file, so the whole thing is the text field
            first_blank_line_index = -1
        else:
            first_blank_line_index = next(idx
                                        for idx, value in enumerate(tiddler_lines)
                                        if not value.strip())
        with tiddler_path.open("w") as f:
            f.writelines(tiddler_lines[0:first_blank_line_index+1])
            f.write(new_text)
    _set_fields(mappings, editor)


def _set_nontext_field(field: str, mappings: Dict[str, str]) -> None:
    "Set the specified /field/ to a new value in each pair of tiddler name/value."
    def editor(tiddler_path: Path, tiddler_lines: Sequence[str], new_text: str) -> None:
        # First check if the field exists and is currently not set to new_text...
        line_to_change = -1
        for idx, line in enumerate(tiddler_lines):
            m = re.match(f"^{field}: (?P<value>.*)", line)
            if m:
                if m.group('value') != new_text:
                    line_to_change = idx
                break

        # ...and if so, write the file again with the change in place.
        if line_to_change != -1:
            with tiddler_path.open("w") as f:
                for idx, line in enumerate(tiddler_lines):
                    if idx == line_to_change:
                        f.write(f"{field}: {new_text}\n")
                    else:
                        f.write(line)
    _set_fields(mappings, editor)


@tzk_builder
def set_tiddler_values(text: Optional[Dict[str, str]] = None,
                       **kwargs: Dict[str, str]) -> None:
    """
    Set fields of selected config or other tiddlers to arbitrary new values.

    This can be used to make customizations that can't easily be done with feature
    flags or other wikitext solutions within the wiki -- for instance, changing the
    subtitle or what buttons are visible. It's also used to implement feature flags
    in the first place by changing the ``$:/config/sib/CurrentEditionPublicity``
    tiddler to ``public``, so at minimum, the build of a public wiki should use:

    .. code-block:: python

        builders.set_tiddler_values({
            '$__config_sib_CurrentEditionPublicity.tid': 'public',
        })

    Any number of arguments can be provided. The name of each argument is the name of
    the field to edit. One positional argument may be used with no name; this argument
    is assumed to be mappings for replacing the 'text' field.
    """
    assert 'public_wiki_folder' in build_state
    for field_name, field_mapping in kwargs.items():
        if not isinstance(field_mapping, Mapping):
            raise AssertionError(
                "Arguments to set_tiddler_values must be dictionaries "
                "mapping tiddler names to values.")

    if text is not None:
        _set_text_field(text)
    
    for field, mappings in kwargs.items():
        _set_nontext_field(field, mappings)


@tzk_builder
def delete_tiddlers(tiddlers: Sequence[str]) -> None:
    """
    Delete selected tiddlers from the output.

    This is hopefully self-explanatory.

    :param tiddlers: A list of filenames of tiddlers to delete.
    """
    assert 'public_wiki_folder' in build_state
    for tiddler in tiddlers:
        tiddler_path = (Path(build_state['public_wiki_folder']) / "tiddlers" / tiddler)
        tiddler_path.unlink()


@tzk_builder
def publish_wiki_to_github(
        output_folder: str = "output/public_site/",
        commit_message: str = "publish checkpoint",
        remote: str = "origin",
        refspec: str = "master",
        push = True) -> None:
    """
    Publish the built wiki to GitHub.

    :param output_folder:  Path to a folder containing the Git repository you'd like 
                           to publish, relative to your private wiki's root directory.
                           This folder should contain the version of your wiki
                           built by the :func:`compile_html_file()` builder,
                           either directly or in a subfolder somewhere.
                           You need to clone or create the repository yourself
                           and set up the remotes as appropriate
                           before running this step.
                           Default ``output/public_site``.
    :param commit_message: Message to use when committing the newly built wiki.
                           Note that all changes in the repository will be committed.
                           Default ``publish checkpoint``.
    :param remote:         The repository remote to push changes to.
                           If you don't know what that is,
                           the default of ``origin`` is probably correct.
    :param refspec:        The local branch or refspec to push.
                           Default ``master``.
    :param push:           If set to False, don't push after committing the changes
                           (mostly useful for testing purposes).
                           Default True.
    """
    with pushd(output_folder):
        if not os.path.isdir(".git"):
            stop(f"The output folder {output_folder} isn't a Git repository. "
                 f"Please go initialize it and then try again.")

        git.exec("add", "-A")
        rc = git.rc("commit", "-m", commit_message)
        if rc == 0:
            if push:
                git.exec("push", remote, refspec)
        elif rc == 1:
            info("No changes to commit or publish. "
                "You probably rebuilt without changing the wiki in between.")
        else:
            stop(f"'git commit' returned unknown return code {rc}.")


@tzk_builder
def shell(shell_command: str) -> None:
    """
    Run an arbitrary shell command.

    The builder will fail if a return code other than 0 is returned,
    otherwise it will succeed. Output will be printed to stdout.

    :param shell_command: A string to be passed to your system's shell.
    """
    info("$ " + shell_command)
    try:
        output = subprocess.check_output(shell_command, shell=True, text=True)
    except subprocess.CalledProcessError as e:
        if e.output.strip():
            stop(f"Command exited with return code {e.returncode}:\n{e.output}")
        else:
            stop(f"Command exited with return code {e.returncode} (no output).")
    else:
        if output.strip():
            info(f"Command exited with return code 0:\n{output}")
        else:
            info(f"Command exited with return code 0 (no output).")


@tzk_builder
def editionify(target_folder: str, description: str) -> None:
    """
    Copy the output folder to a target location and set its edition description.

    This generates a TiddlyWiki edition based on the temporary output. By
    copying it into an appropriate location or setting the environment variable
    :envvar:`TIDDLYWIKI_EDITION_PATH` to the parent directory of the edition's folder,
    it's possible to quickly generate new TiddlyWikis based on the edition
    "template" (``tiddlywiki --init EDITION_NAME``, where the ``EDITION_NAME`` is the
    name of the folder).

    :param target_folder: The folder to copy the output folder to.
                          This folder *will be deleted* if it already exists
                          prior to the copy.
    :param description:   The description of this edition to use in the new edition's
                          ``tiddlywiki.info`` file.
    """
    try:
        shutil.rmtree(target_folder)
    except FileNotFoundError:
        pass
    shutil.copytree(
        build_state['public_wiki_folder'],
        target_folder,
    )

    def editor(tinfo):
        tinfo['description'] = description
        return tinfo
    alter_tiddlywiki_info(Path(target_folder) / "tiddlywiki.info", editor)


@tzk_builder
def add_plugins(plugins: Sequence[str]) -> None:
    """
    Add one or more plugins to the tiddlywiki.info file.

    :param plugins: A list of plugin names (e.g., "tiddlywiki/codemirror") to add.
    """
    def editor(tinfo):
        tinfo['plugins'].extend(plugins)
        return tinfo
    alter_tiddlywiki_info(Path(build_state['public_wiki_folder']) / "tiddlywiki.info",
                          editor)



"""
config.py - read and manage the tzk config file
"""
import datetime
import functools
import importlib
import os
from pathlib import Path
import sys
from typing import Any

from tzk.util import fail


DEFAULT_INIT_OPTS = {
    'wiki_name': 'wiki',
    'tw_version_spec': '^5.1.23',
    'author': None,
}


class ConfigurationManager:
    def __init__(self):
        self.initialize_cm()

    def __getattr__(self, attr):
        if self.conf_mod is None:
            return None
        else:
            return getattr(self.conf_mod, attr, None)

    def initialize_cm(self):
        self.config_path = Path.cwd()

        for child in sorted(self.config_path.iterdir()):
            if child.is_file() and child.name.endswith('.py'):
                mod_name = child.name.rsplit('.', 1)[0]
                if mod_name == 'tzk_config':
                    sys.path.insert(0, Path("__file__").parent)
                    sys.path.insert(0, str(self.config_path))
                    self.conf_mod = importlib.import_module(mod_name)
                    del sys.path[0:1]
                    break
        else:
            # no config file
            self.conf_mod = None

    def has_config(self) -> bool:
        return self.conf_mod is not None

    def require_config(self) -> None:
        """
        Quit with exit status 1 if no config file was found.
        """
        if not self.has_config():
            fail(f"No tzk_config.py found in the current directory. "
                 f"(Try 'tzk init' if you want to create a new one.)")


def cm(cache=[]):
    """
    Call this function to retrieve the singleton ConfigurationManager object,
    reading and initializing it if necessary.

    Since so much happens when the ConfigurationManager is initialized,
    this has to go in a function so that autodoc doesn't blow up
    when it tries to import the module.
    """
    if not cache:
        cache.append(ConfigurationManager())
    return cache[0]



"""
setup.py -  setuptools configuration for esc
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tzk",
    version="0.1.2",
    author="Soren I. Bjornstad",
    author_email="zettelkasten@sorenbjornstad.com",
    description="Build tool for TiddlyWiki Zettelkasten",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sobjornstad/tzk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "tzk = tzk.__main__:launch"
        ],
    },
    python_requires='>=3.6',
    include_package_data=True,
)



from abc import ABC, abstractmethod, abstractclassmethod
import argparse
import os
from pathlib import Path
import shutil
import sys
import traceback
from typing import Optional

from tzk.config import cm, DEFAULT_INIT_OPTS
from tzk import git
from tzk import tw
from tzk.util import (BuildError, fail, numerize, require_dependencies, pushd,
                      TZK_VERSION)


class CliCommand(ABC):
    """
    Base class for subcommands of tzk.
    """
    #: The text of the subcommand to be used on the command line.
    cmd = None   # type: str
    #: Help string for argparse to display for this subcommand.
    help = None  # type: str

    @abstractclassmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        """
        Given the :arg:`parser`, add any arguments this subcommand wants to accept.
        """
        raise NotImplementedError

    @abstractmethod
    def execute(self, args: argparse.Namespace) -> None:
        """
        Given the :arg:`args` passed to this subcommand, do whatever the command does.
        """
        raise NotImplementedError


class CommitCommand(CliCommand):
    cmd = "commit"
    help = "Commit all changes to the wiki repository."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-m", "--message",
            metavar="MSG",
            help="Commit message to use.",
            default=cm().commit_message,
        )
        parser.add_argument(
            "-r", "--remote",
            metavar="REMOTE",
            help="Name of the Git remote to push to.",
            default=cm().commit_remote,
        )
        parser.add_argument(
            "-l", "--local",
            help="Don't push the results to any configured remote repository.",
            action="store_true",
        )

    def execute(self, args: argparse.Namespace) -> None:
        cm().require_config()
        chdir_to_wiki()
        if cm().commit_require_branch:
            current_branch = git.read("rev-parse", "--abbrev-ref", "HEAD")
            if current_branch != cm().commit_require_branch:
                fail(f"You are on the '{current_branch}' branch, "
                     f"but your tzk configuration requires you to be on the "
                     f"'{cm().commit_require_branch}' branch to commit.")

        git.exec("add", "-A")
        if git.rc("commit", "-m", args.message) == 0 and args.remote and not args.local:
            git.exec("push", args.remote)


class ListenCommand(CliCommand):
    cmd = "listen"
    help = "Start a TiddlyWiki server in the current tzk repository."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-p", "--port",
            metavar="PORT",
            help="Port to listen on.",
            default=str(cm().listen_port or "8080"),
        )
        parser.add_argument(
            "--username",
            metavar="USERNAME",
            default=cm().listen_username or "",
            help="Username to use for basic authentication, if any.",
        )
        parser.add_argument(
            "--password",
            metavar="PASSWORD",
            default=cm().listen_password or "",
            help="Password to use for basic authentication, if any.",
        )
    
    def execute(self, args: argparse.Namespace) -> None:
        cm().require_config()
        chdir_to_wiki()
        try:
            tw.exec(
                [
                    ("listen",
                    f"port={args.port}",
                    f"username={args.username}",
                    f"password={args.password}")
                ]
            )
        except KeyboardInterrupt:
            # We'll terminate anyway now that we're at the end of execute() --
            # no need to display the traceback to the user.
            pass


class InitCommand(CliCommand):
    cmd = "init"
    help = "Set up a new tzk repository."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-n", "--wiki-name",
            metavar="NAME",
            help="The wiki will be installed in a subfolder of the current directory, called NAME.",
            default=DEFAULT_INIT_OPTS['wiki_name'],
        )
        parser.add_argument(
            "-v", "--tiddlywiki-version-spec",
            metavar="SPEC",
            help="NPM version spec for the version of TiddlyWiki to start your package.json at.",
            default=DEFAULT_INIT_OPTS['tw_version_spec'],
        )
        parser.add_argument(
            "-a", "--author",
            metavar="AUTHOR_NAME",
            help="The author to be credited in the package.json, if not your current username.",
            default=DEFAULT_INIT_OPTS['author'],
        )

    def _precheck(self):
        require_dependencies()

        if os.path.exists("package.json"):
            fail("A 'package.json' file already exists in the current directory. "
                 "Perhaps you've already initialized a tzk repository here?")

    def execute(self, args: argparse.Namespace) -> None:
        self._precheck()
        tw.install(args.wiki_name, args.tiddlywiki_version_spec, args.author)


class PreflightCommand(CliCommand):
    cmd = "preflight"
    help = "Check if tzk and all its dependencies are correctly installed."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        pass

    def execute(self, args: argparse.Namespace) -> None:
        require_dependencies()
        print("You're all set! Change into a directory you want to "
              "turn into your tzk repository and run 'tzk init'.")


class VersionCommand(CliCommand):
    cmd = "version"
    help = "Find out what version your tzk is at."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        pass

    def execute(self, args: argparse.Namespace) -> None:
        print(f"tzk version {TZK_VERSION}")


class BuildCommand(CliCommand):
    cmd = "build"
    help = ("Build another wiki or derivative product, "
            "such as a public version of the wiki, "
            "from this tzk repository.")

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "product",
            metavar="PRODUCT",
            help="Name of the product you want to build (defined in your config file).",
        )
        parser.add_argument(
            "-s", "--skip-builder",
            metavar="BUILDER_NAME",
            help="Function name of a builder to skip even if part of the PRODUCT. "
                 "This option can be specified multiple times.",
            action="append",
            default=[],
        )

    def _precheck(self, product: str) -> None:
        if cm().products is None:
            fail("No 'products' dictionary is defined in your config file.")
        if product not in cm().products:
            fail(f"No '{product}' product found in the products dictionary "
                 f"in your config file. (Available: {', '.join(cm().products.keys())})")
        if not cm().products[product]:
            fail(f"No build steps are defined in the '{product}' product "
                 f"in your config file.")

    def execute(self, args: argparse.Namespace) -> None:
        cm().require_config()
        chdir_to_wiki()
        self._precheck(args.product)

        # Find the build steps for the product the user specified.
        steps = cm().products[args.product]
        print(f"tzk: Starting build of product '{args.product}'.")
        print(f"tzk: Found {len(steps)} build {numerize(len(steps), 'step')}.")

        failed = False
        # For each build step...
        try:
            for idx, step in enumerate(steps, 1):
                # Explain what we're doing. Use first line of the builder's docstring
                # as a summary, if present.
                if hasattr(step, '__doc__') and step.__doc__ is not None:
                    short_description = step.__doc__.strip().split('\n')[0].rstrip('.')
                    print(f"tzk: Step {idx}/{len(steps)}: {short_description}")
                else:
                    print(f"tzk: Step {idx}/{len(steps)}")

                # If the user asked to skip this builder on the command line, do so.
                if step.__name__ in args.skip_builder:
                    print(f"tzk: Skipping step {idx} due to --skip-builder parameter.")
                    continue

                # Execute step and handle any errors.
                step()

        except BuildError as e:
            failed = True
            print(f"tzk: ERROR: {str(e)}")
            print(f"tzk: Build of product '{args.product}' failed on step {idx}, "
                  f"backed by builder '{step.__name__}'. ")
            print(f"tzk: Add '--skip-builder {step.__name__}' if you'd like "
                  f"to skip this step.")

        except Exception:
            failed = True
            print(f"tzk: Build of product '{args.product}' failed on step {idx}: "
                  f"unhandled exception. "
                  f"The original error follows:")
            traceback.print_exc()

        finally:
            for idx, step in enumerate(steps, 1):
                if hasattr(step, 'cleaner'):
                    print(f"tzk: Running cleanup routine registered by step {idx}...")
                    step.cleaner()

        if failed:
            sys.exit(1)
        else:
            print(f"tzk: Build of product '{args.product}' completed successfully.")


class ConvertCommand(CliCommand):
    cmd = "convert"
    help = "Convert a tzk repository to a single-file wiki or vice versa."

    @classmethod
    def setup_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "source",
            metavar="SOURCE",
            help="Wiki to convert. "
                 "Either a folder containing a tzk_config.py file or an HTML file.",
        )
        parser.add_argument(
            "destination",
            metavar="DEST",
            help="Output location for the converted wiki. "
                 "Either a folder (existing or not) or the name of an HTML file."
        )
        parser.add_argument(
            "-f", "--force",
            action="store_true",
            help="Overwrite the destination location if it already exists.",
        )

    def _precheck(self, args: argparse.Namespace) -> None:
        if not os.path.exists(args.source):
            fail(f"The source location '{args.source}' does not exist.")
        if os.path.exists(args.destination) and not args.force:
            fail(f"The destination location '{args.destination}' already exists. "
                 f"(Use --force to overwrite it.)")


    def execute(self, args: argparse.Namespace) -> None:
        require_dependencies()
        self._precheck(args)

        source = Path(args.source).absolute()
        destination = Path(args.destination)
        source_type = 'file' if source.is_file() else 'folder'
        dest_type = 'file' if destination.name.endswith(".html") else 'folder'

        # If types are the same, there's nothing to convert...
        if source_type == dest_type:
            fail("The source or the destination may not be of the same type. "
                 "One must be a folder and the other an HTML file.")

        # Conversion from folder to file using --render.
        if source_type == 'folder':
            if not Path(source / "tzk_config.py").exists():
                fail("The source folder '{source}' does not contain "
                     "a tzk_config.py file.")

            with pushd(str(source)):
                source_wiki_folder = Path.cwd() / cm(cache=[]).wiki_folder

            tw.exec(
                (
                    ("output", str(destination.parent)),
                    ("render", "$:/core/save/all", destination.name, "text/plain"),
                ),
                base_wiki_folder=str(source_wiki_folder)
            )

        # Conversion from file to folder using --savewikifolder.
        elif source_type == 'file':
            if destination.exists() and args.force:
                doing_what = "Overwriting existing"
                shutil.rmtree(destination)
            else:
                doing_what = "Creating new"
            print(f"tzk: {doing_what} tzk repository in destination '{destination}'...")
            os.mkdir(destination)

            def installer(wiki_name):
                tw.exec(
                    (
                        ("load", str(source)),
                        ("savewikifolder", wiki_name),
                    ),
                )
            with pushd(str(destination)):
                tw.install(_tw_func=installer, **DEFAULT_INIT_OPTS)  # type: ignore

        else:
            raise AssertionError(f"Invalid source type {source_type}.")


def chdir_to_wiki():
    """
    For most operations, we want the current directory to be the wiki folder.
    """
    if not cm().wiki_folder:
        fail("No 'wiki_folder' option found in config. Set this option to the name "
            "of the wiki subfolder within the current directory.")

    try:
        os.chdir(cm().wiki_folder)
    except FileNotFoundError:
        fail(f"Tried to change directory into the wiki_folder '{cm().wiki_folder}' "
             f"specified in your config file, but that directory does not exist.")

    if not os.path.exists("tiddlywiki.info"):
        fail(f"After changing directory into {cm().wiki_folder} per your config file: "
             f"Expected a 'tiddlywiki.info' file in {os.getcwd()}. "
             f"Please check that your wiki is initialized "
             f"and you specified the correct wiki_folder_name.")


def launch():
    # If this isn't a tzk directory but we have an environment variable listing one,
    # go there before doing anything else.
    if (not os.path.exists("tzk_config.py")
            and os.environ.get('TZK_DIRECTORY')
            and sys.argv[1] != "init"):  # we can't init an existing TZK_DIRECTORY
        try:
            os.chdir(os.environ['TZK_DIRECTORY'])
        except (FileNotFoundError, NotADirectoryError):
            fail(f"The current directory contains no 'tzk_config.py' "
                 f"and the TZK_DIRECTORY environment variable is set to "
                 f"'{os.environ['TZK_DIRECTORY']}', so I tried to go there, "
                 f"but it doesn't exist or is not a directory.")
        if not os.path.exists("tzk_config.py"):
            fail(f"The current directory contains no 'tzk_config.py' "
                 f"and the TZK_DIRECTORY environment variable is set to "
                 f"'{os.environ['TZK_DIRECTORY']}', so I tried to go there, "
                 f"but there's no tzk_config.py there either.")

    parser = argparse.ArgumentParser(
        description=f"TiddlyZettelKasten {TZK_VERSION} CLI\n"
                    f"Copyright (c) 2021 Soren Bjornstad.\n"
                    f"MIT license; see https://github.com/sobjornstad/tzk/blob/master/LICENSE for details.",
        epilog="For full documentation, see https://tzk.readthedocs.io/en/latest/.",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers()
    for command in sorted(CliCommand.__subclasses__(), key=lambda i: i.__name__):
        subparser = subparsers.add_parser(command.cmd,
                                          help=command.help,
                                          description=command.help)
        subparser.set_defaults(_cls=command)
        command.setup_arguments(subparser)  # type: ignore

    args = parser.parse_args()
    if not hasattr(args, '_cls'):
        # no subcommand was given
        parser.print_help()
        sys.exit(0)

    args._cls().execute(args)


if __name__ == '__main__':
    launch()



import datetime
import functools
import json
import os
from pathlib import Path
import subprocess
from textwrap import dedent
from typing import Callable, Optional, Sequence

from tzk import config
from tzk import git
from tzk.util import pushd


@functools.lru_cache(1)
def _npm_bin() -> str:
    return subprocess.check_output(("npm", "bin"), text=True).strip()


def _tw_path() -> str:
    return _npm_bin() + "/tiddlywiki"


@functools.lru_cache(1)
def _whoami() -> str:
    "Try to guess the user's name."
    try:
        return subprocess.check_output(("whoami",), text=True).strip()
    except subprocess.CalledProcessError:
        return "user"


def exec(args: Sequence[Sequence[str]], base_wiki_folder: str = None) -> int:
    """
    Execute a series of TiddlyWiki commands.

    :param args: A list of lists of CLI commands to send to TiddlyWiki.
                 The first element of each list is a TiddlyWiki CLI command,
                 without the ``--``, e.g., ``savewikifolder``.
                 The following elements of the list are arguments to that command.
    :param base_wiki_folder: If the wiki to execute commands against is not the one
                             in the current directory, provide its path here.
                             The current directory is the source wiki's root directory
                             during the execution of builders,
                             unless explicitly changed.
    """
    # must pushd into base wiki to find the tiddlywiki node_modules
    if base_wiki_folder is not None:
        with pushd(base_wiki_folder):
            call_args = [_tw_path()]
    else:
        call_args = [_tw_path()]

    if base_wiki_folder is not None:
        call_args.append(base_wiki_folder)
    for tw_arg in args:
        call_args.append(f"--{tw_arg[0]}")
        for inner_arg in tw_arg[1:]:
            call_args.append(inner_arg)
    return subprocess.call(call_args)


def _init_tzk_config() -> None:
    print("tzk: Creating new tzk_config.py...")
    with open(Path(__file__).parent / "default_config.py") as f:
        default_config = f.read()

    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    generation_details = f"Created automatically by 'tzk init' at {now}."
    default_config = default_config.replace('<<GENERATION_DETAILS>>',
                                            generation_details)

    with open("tzk_config.py", "w") as f:
        f.write(default_config)


def _init_npm(wiki_name: str, tw_version_spec: str, author: str) -> None:
    """
    Create a package.json file for this repository, requiring TiddlyWiki
    at the specified version, and install the npm dependencies.
    """
    print("tzk: Creating new package.json...")
    PACKAGE_JSON = dedent("""
    {
        "name": "%(wiki_name)s",
        "version": "1.0.0",
        "description": "My nice notes",
        "dependencies": {
            "tiddlywiki": "%(tw_version_spec)s"
        },
        "author": "%(author)s",
        "license": "See copyright notice in wiki"
    }
    """).strip() % ({'tw_version_spec': tw_version_spec, 'author': author,
                     'wiki_name': wiki_name})
    with open("package.json", "w") as f:
        f.write(PACKAGE_JSON)

    print("tzk: Installing npm packages from package.json...")
    subprocess.check_call(("npm", "install"))


def _init_tw(wiki_name: str) -> None:
    """
    Create a new TiddlyWiki in the subfolder named 'wiki_name'
    using 'tiddlywiki --init'.
    """
    print("tzk: Creating new TiddlyWiki...")
    try:
        os.mkdir(wiki_name)
    except FileExistsError:
        pass
    with pushd(wiki_name):
        old_edition_path = os.environ.get('TIDDLYWIKI_EDITION_PATH')
        os.environ['TIDDLYWIKI_EDITION_PATH'] = str(Path(__file__).parent / "editions")
        try:
            subprocess.check_call((_tw_path(), "--init", "tzk"))
        finally:
            if old_edition_path:
                os.environ['TIDDLYWIKI_EDITION_PATH'] = old_edition_path


def _restore_plugins(wiki_name: str) -> None:
    """
    Add the two plugins required for client-server operation to the existing
    ones in the edition, sort them in order, and replace the existing plugins
    array in the tiddlywiki.info.
    """
    print("tzk: Configuring plugins in tiddlywiki.info...")

    info_path = Path.cwd() / wiki_name / "tiddlywiki.info"
    edition_path = Path(__file__).parent / "editions" / "tzk" / "tiddlywiki.info"

    with info_path.open("r") as f:
        info_data = json.load(f)
    with edition_path.open("r") as f:
        edition_data = json.load(f)

    plugins = {"tiddlywiki/filesystem", "tiddlywiki/tiddlyweb"}
    plugins = plugins.union(edition_data['plugins'])
    info_data['plugins'] = sorted(plugins)

    with info_path.open("w") as f:
        json.dump(info_data, f, indent=4)


def _init_gitignore() -> None:
    """
    Create a basic gitignore for the new wiki.
    """
    print("tzk: Creating gitignore...")
    GITIGNORE = dedent("""
    __pycache__/
    node_modules/
    .peru/
    output/

    \$__StoryList.tid
    """).strip()
    with open(".gitignore", "w") as f:
        f.write(GITIGNORE)


def _initial_commit() -> None:
    """
    Create a new Git repo and commit everything we've done so far.
    """
    print("tzk: Initializing new Git repository for wiki...")
    git.exec("init")

    print("tzk: Committing changes to repository...")
    git.exec("add", "-A")
    output = git.read("commit", "-m", "Initial commit")
    # Print just a summary since there are going to be a lot of files.
    print('\n'.join(output.split('\n')[0:2]))


def install(wiki_name: str, tw_version_spec: str, author: Optional[str],
            _tw_func: Optional[Callable[[str], None]] = None):
    """
    Install TiddlyWiki on Node.js in the current directory and set up a new wiki.

    If _tw_func is provided, call it to create the TiddlyWiki in the new tzk repository
    rather than the default routine. It receives one argument, the name of the new wiki.
    """
    # assert: caller has checked npm and git are installed

    if author is None:
        author = _whoami()

    _init_tzk_config()
    _init_npm(wiki_name, tw_version_spec, author)

    if _tw_func is not None:
        _tw_func(wiki_name)
    else:
        _init_tw(wiki_name)

    _restore_plugins(wiki_name)
    _init_gitignore()
    _initial_commit()

    print("tzk: Initialized successfully. "
          "Review the 'tzk_config.py' in a text editor and make any changes desired, "
          "then run 'tzk listen' to start the server.")
